from concurrent.futures import Executor, as_completed
from logging import Logger
from time import time
import traceback
from typing import Any, Dict, Iterable, List, Optional, Union
from boto3.session import Session
import requests

import orjson


def partial_send_sqs(url, **kw):
    def send(payload: str, **kws):
        return (
            Session()
            .client("sqs")
            .send_message(QueueUrl=url, MessageBody=payload, **{**kw, **kws})
        )

    return send


TASK_TOKEN = "taskToken"


def send_task_heartbeat(sID: str, out: List[List[Dict[str, Any]]], logger: Logger):
    if (
        isinstance(out, Iterable)
        and (len(out) > 0)
        and (len(out[-1]) > 0)
        and (TASK_TOKEN in out[-1][0])
    ):
        return logger.info(
            "send task heartbeat",
            sID=sID,
            token=out[TASK_TOKEN],
            res=Session()
            .client("stepfunctions")
            .send_task_heartbeat(taskToken=out[TASK_TOKEN]),
        )


def send_task_success(sID: str, result, out: Dict[str, Any], logger: Logger):
    if TASK_TOKEN in out:
        return logger.info(
            "send task success",
            sID=sID,
            token=out[TASK_TOKEN],
            res=Session()
            .client("stepfunctions")
            .send_task_success(taskToken=out[TASK_TOKEN], output=result),
        )


def send_task_failure(
    sID: str, exception: Exception, out: List[List[Dict[str, Any]]], logger: Logger
):
    if (
        isinstance(out, Iterable)
        and (len(out) > 0)
        and (len(out[-1]) > 0)
        and (TASK_TOKEN in out[-1])
    ):
        return logger.info(
            "send task failure",
            sID=sID,
            token=out[TASK_TOKEN],
            res=Session()
            .client("stepfunctions")
            .send_task_failure(
                taskToken=out[TASK_TOKEN],
                error=str(exception),
                cause=traceback.format_exc(),
            ),
        )


URL = SQS_URL = "url"
SID_FORMAT = "{sid}"


def formatSubSQS(
    route: Dict[str, Any], sID: str, TTL: Optional[float] = None
) -> Dict[str, Any]:
    if (SQS_URL in route) and (SID_FORMAT in route[SQS_URL]):
        # URL https://xxxxxxxxxxxxxxxx/yyyyyyyyy#{sid}
        if TTL is None:
            TTL = time()
        route[SQS_URL] = route[SQS_URL].format(sid=sID, TTL=TTL)
    return route


def SQS(sID: str, result, out: Dict[str, Any], logger: Logger):
    if SQS_URL in out:
        return logger.info(
            "send SQS success",
            sID=sID,
            sqs=out[SQS_URL],
            res=Session()
            .client("sqs")
            .send_message(QueueUrl=out[SQS_URL], MessageBody=result),
        )


SNS_TOPIC_ARN = "arn"
SNS_TOPIC = "topic"


def SNS(sID: str, result, out: Dict[str, Any], logger: Logger):
    if SNS_TOPIC_ARN in out:
        return logger.info(
            "publish SNS success",
            sID=sID,
            token=out[TASK_TOKEN],
            res=Session()
            .client("sns")
            .publish(
                TopicArn=out[SNS_TOPIC_ARN],
                Message=result,
                Subject=out.get(SNS_TOPIC, out[SNS_TOPIC_ARN]),
            ),
        )


LAMBDA_FUNC = "arn"
INVOKE_SYNC = "sync"


def Lambda(sID: str, result, out: Dict[str, Any], logger: Logger) -> Union[Dict, None]:
    if LAMBDA_FUNC in out:
        if out.get(INVOKE_SYNC):
            return (
                Session()
                .client("lambda")
                .invoke(
                    FunctionName=out[LAMBDA_FUNC],
                    InvocationType="RequestResponse",
                    LogType="tail",
                    Payload=result,
                )
            )

        return logger.info(
            "publish SNS success",
            sID=sID,
            token=out[TASK_TOKEN],
            res=Session()
            .client("lambda")
            .invoke(
                FunctionName=out[LAMBDA_FUNC], InvocationType="Event", Payload=result,
            ),
        )


HTTP_METHOD = "method"


def HTTP(sID: str, result, out: Dict[str, Any], logger: Logger):
    if URL in out:
        return logger.info(
            "HTTP GET", sID=sID, url=out[URL], res=requests.get(out[URL])
        )


def default_invoker(*arg, **kw):
    pass


CASE = {
    "aws:state": send_task_success,
    "aws:sqs": SQS,
    "aws:sns": SNS,
    "aws:lambda": Lambda,
    "http": HTTP,
    None: default_invoker,
}

OUT = "out"
SERVICE = "srv"


def invoker(
    sID: str, result: Dict[str, Any], logger: Logger, pool: Optional[Executor] = None,
):
    if (OUT not in result) or (len(result[OUT]) <= 0):
        return
    out: List[Dict[str, Any]] = result[OUT].pop(0)
    res = str(orjson.dumps(result, option=orjson.OPT_SERIALIZE_NUMPY), "utf-8")

    if pool is None:
        return [
            CASE.get(i.get(SERVICE), default_invoker)(
                sID=sID, result=res, out=i, logger=logger
            )
            for i in out
        ]

    else:
        return [
            i.result()
            for i in as_completed(
                [
                    pool.submit(
                        CASE.get(i.get(SERVICE), default_invoker),
                        sID=sID,
                        result=res,
                        out=i,
                        logger=logger,
                    )
                    for i in out
                ]
            )
        ]


"""{
    OUT: [
        [{SERVICE: "aws:sns", SNS_TOPIC_ARN: ""}],
        [{SERVICE: "aws:sqs", SQS_URL: ""}, {SERVICE: "http", URL: ""}],
        [{SERVICE: "aws:state", TASK_TOKEN: ""}],
    ]
}"""
