from concurrent.futures import Executor, ThreadPoolExecutor, ProcessPoolExecutor, Future
import gc
from logging import Logger, getLogger
from multiprocessing import Pipe, Process, freeze_support
from multiprocessing.connection import Connection
from os import cpu_count
from time import sleep
from typing import Any, Callable, Iterable, List, Mapping, Optional, Tuple, Union

gc.enable()


class MapFuture:
    futures: List[Future]

    def __init__(self) -> None:
        super().__init__()
        self.futures = []

    def result(self, timeout: Union[float, None] = None) -> Iterable:
        if len(self.futures) <= 0:
            return []

        return [i.result(timeout) for i in self.futures]

    def done(self, agg: Callable[[Iterable], Any] = all) -> Any:
        """gather done of Futures

        Args:
            agg (Callable[[Iterable], Any], optional): Aggregate function i.e. all, any, sum, list. Defaults to all.

        Returns:
            Any: Depend on agg function
        """
        return agg(i.done() for i in self.futures)


class AWSLambdaMultiPoolExecutor(ThreadPoolExecutor):
    def __init__(
        self,
        max_workers: Union[int, None] = None,
        thread_name_prefix: Optional[str] = None,
        initializer: Optional[Callable[..., None]] = None,
        initargs: Optional[Tuple[Any, ...]] = None,
        logger: Optional[Logger] = None,
    ) -> None:
        super().__init__(max_workers, thread_name_prefix, initializer, initargs)
        self.logger = getLogger() if (logger is None) else logger

    @classmethod
    def _worker(cls, fn: Callable, logger: Logger, *arg, **kw):
        main, sub = Pipe()
        proc = Process(
            target=cls.process_worker, args=[fn, sub, logger] + list(arg), kwargs=kw
        )
        proc.start()
        while (not main.poll(None)) and proc.is_alive():
            sleep(0.001)

        status, result = main.recv()
        if status:
            return result
        else:
            raise result from result

    def submit_process(self, fn: Callable, *arg, **kw) -> Future:
        return super().submit(self._worker, fn, self.logger, *arg, **kw)

    @staticmethod
    def process_worker(fn: Callable, pipe: Connection, logger: Logger, *arg, **kw):
        try:
            result = fn(*arg, **kw)
            pipe.send([True, result])
        except Exception as e:
            logger.exception("{e} from {fn}(*{arg},**{kw})", e=e, fn=fn, arg=arg, kw=kw)
            pipe.send([False, e])
        return gc.collect()


class Pool:
    def __init__(
        self,
        process_pool: Union[None, bool, ThreadPoolExecutor, ProcessPoolExecutor] = None,
        pool_worker: Optional[int] = None,
        aws_lambda: Optional[bool] = None,
    ) -> None:

        freeze_support()
        cpu = cpu_count()
        if process_pool is None:
            if pool_worker == 0:
                self.pool = None
                return
            process_pool = cpu > 1

        if isinstance(process_pool, bool):
            if process_pool:
                if (pool_worker is None) or (pool_worker == 0):
                    pool_worker = cpu
                elif pool_worker < 0:
                    pool_worker = max(1, pool_worker - cpu_count())

                try:
                    if aws_lambda:
                        self.pool = AWSLambdaMultiPoolExecutor(max_worker=pool_worker)
                    else:
                        self.pool = ProcessPoolExecutor(max_workers=pool_worker)
                except OSError:
                    self.pool = AWSLambdaMultiPoolExecutor(max_worker=pool_worker)
            else:
                if (pool_worker is None) or (pool_worker == 0):
                    pool_worker = min(1 + (cpu * 4), 61)
                elif pool_worker < 0:
                    pool_worker = min(max(1, (4 * cpu) + 1 - pool_worker), 61)

                self.pool = ThreadPoolExecutor(max_workers=pool_worker)

        else:
            self.pool = process_pool

    def __del__(self):
        if self.pool is None:
            self.pool.shutdown(wait=False)

    def submit(self, fn: Callable, *arg, **kw) -> Future:
        return self.pool.submit(fn, *arg, **kw)

    def automap_async(self, fn: Callable, args_kw: Iterable):
        futures = MapFuture()

        for i in args_kw:
            if isinstance(i, Mapping):
                futures.futures.append(self.pool.submit(fn, **i))
            elif isinstance(i, Iterable):
                futures.futures.append(self.pool.submit(fn, *i))
            else:
                futures.futures.append(self.pool.submit(fn, i))

        return futures

    def map_async(self, fn: Callable, args_kw: Iterable):
        futures = MapFuture()

        for i in args_kw:
            futures.futures.append(self.pool.submit(fn, i))

        return futures


def orchastrator(
    fn_pool: Iterable[Tuple[Callable, Union[Pool, Executor]]],
    logger: Logger,
    break_if_None: bool = True,
    series_compute: int = 1,
):
    assert len(fn_pool) > 0, "fn_pool should contain at least one step"
    # build orchastrator function

    ## Build series compute
    if series_compute <= 0:

        def series(x):
            return x

    elif break_if_None:

        def series(xs):
            ys = []
            for x in xs:
                y = fn_pool[0][0](x)

                for fn in fn_pool[1:]:
                    if y is None:
                        break
                    y = fn[0](y)
                ys.append(y)
            return ys

    else:

        def series(xs):
            ys = []
            for x in xs:
                y = fn_pool[0][0](x)

                for fn in fn_pool[1:]:
                    y = fn[0](y)
                ys.append(y)
            return ys

    ## Build parallel compute

    if break_if_None:

        def done_callback(x):
            pass
